#include <glad/glad.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <learnopengl/shader.h>
#include <learnopengl/camera.h>

#include <stdio.h>
#include <string>

#include "config.h"

// ͨú

void framebufferSizeCallback(GLFWwindow* window, int width, int height);
void mouseCallback(GLFWwindow* window, double xpos, double ypos);
void scrollCallback(GLFWwindow* window, double xoffset, double yoffset);
void processInput(GLFWwindow *window);
void checkError();

// к

void printInfos();
void printFrameInfo();

GLuint quadVAO, quadVBO;
void renderQuad();

int main(int argc, char** argv) {
    std::string fragmentShader = "shaders/desk.frag";
    // 0. 
    {
        // fragment shader
        {
            if (argc >= 2) {
                fragmentShader = std::string(argv[1]);
            }
            // monster scene
            if (fragmentShader.find("monster") != std::string::npos) {
                camera = Camera(
                    glm::vec3(0.648786f, 0.424139f, -0.0974952f),
                    glm::vec3(0.0, 1.0f, 0),
                    -257.7f, -0.500005f);
                zoomFactor = 1.0f;
            }
            // primitives scene
            else if (fragmentShader.find("primitives") != std::string::npos) {
                camera = Camera(
                    glm::vec3(-2.82f, 2.23f, -4.86f),
                    glm::vec3(0.0, 1.0f, 0),
                    -304.8f, -25.1f);
                zoomFactor = 2.5f;
            }
            // rabbit scene
            else if (fragmentShader.find("rabbit") != std::string::npos) {
                camera = Camera(
                    glm::vec3(-0.550827f, -0.406117f, -6.52196f),
                    glm::vec3(0.0, 1.0f, 0),
                    -265.6f, 11.1f);
                zoomFactor = 1.0f;
            }
            // desk scene
            else if (fragmentShader.find("desk") != std::string::npos) {
                camera = Camera(
                    glm::vec3(0.0729812f, 1.69295f, -2.06917f),
                    glm::vec3(0.0, 1.0f, 0),
                    449.6f, -9.9f);
                zoomFactor = 1.73f;
            }
        }
        if (argc >= 3) {
            SCR_WIDTH *= 2;
            SCR_HEIGHT *= 2;
            lastX *= 2;
            lastY *= 2;
        }
    }
    printInfos();

    // 1. Ϣ
    GLFWwindow* window;
    {
        glfwInit();
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
        glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
#ifdef __APPLE__
        glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif
        window = glfwCreateWindow(SCR_WIDTH, SCR_HEIGHT, "Raymarching SDFs", NULL, NULL);
        if (window == NULL) {
            printf("Failed to create GLFW window\n");
            glfwTerminate();
            return -1;
        }
        glfwMakeContextCurrent(window);
        glfwSetCursorPosCallback(window, mouseCallback);
        glfwSetScrollCallback(window, scrollCallback);
        glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED); // ȫ
        // glfwSetFramebufferSizeCallback(window, framebufferSizeCallback); // ޸ĴС
        // glfwSwapInterval(0); // ٵȴòˢһν
        glfwSetWindowSizeLimits(window, SCR_WIDTH, SCR_HEIGHT, SCR_WIDTH, SCR_HEIGHT);

        if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress)) {
            printf("Failed to initialize GLAD\n");
            return -1;
        }
    }

    //  shader
    Shader shader("shaders/quad.vert", fragmentShader.c_str());

    //  shader е uniform 
    shader.use();
    shader.setVec3("iResolution", (float)SCR_WIDTH, (float)SCR_HEIGHT, 1.0f);

    // OpenGL ״̬
    glEnable(GL_DEPTH_TEST);

    int iFrame = -1;
    startTime = (float)glfwGetTime();

    // Ⱦ
    while (!glfwWindowShouldClose(window)) {
        // Ƿ
        checkError();

        // ʱ
        float currentFrameTime = (float)glfwGetTime() - startTime;
        deltaTime = currentFrameTime - lastFrameTime;
        lastFrameTime = currentFrameTime;
        timeCost += deltaTime;
        FPS = 1.0f / deltaTime;
        ++iFrame;

        // ʾϢ
        printFrameInfo();

        // ¼
        processInput(window);

        // Ⱦ, СΪ [-0.5, 0.5]
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
        //  buffer
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        //  shader 
        shader.use();
        shader.setFloat("iTime", currentFrameTime);
        shader.setInt("iFrame", iFrame);
        shader.setVec4("iMouse", glm::vec4(lastX, lastY, 0.0f, 0.0f));
        shader.setVec3("camera.position", camera.Position);
        shader.setVec3("camera.front", camera.Front);
        shader.setVec3("camera.up", camera.Up);
        shader.setVec3("camera.right", camera.Right);
        shader.setFloat("camera.focalLength", zoomFactor * 45.0f / camera.Zoom);
        shader.setVec3("lightPositionOffset", lightPosition);
        // Ⱦ
        renderQuad();

        // 
        glfwSwapBuffers(window);
        glfwPollEvents();
    }

    // delete
    // VAO, VBO
    glDeleteVertexArrays(1, &quadVAO);
    glDeleteBuffers(1, &quadVBO);

    checkError();

    glfwDestroyWindow(window);
    glfwTerminate();
    return 0;
}

// Ⱦ [-1, 1]^2
void renderQuad() {
    if (quadVAO == 0) {
        float quadVertices[] = {
            // positions
            -1.0f,  1.0f, 0.0f,
            -1.0f, -1.0f, 0.0f,
             1.0f,  1.0f, 0.0f,
             1.0f, -1.0f, 0.0f,
        };
        // setup plane VAO
        glGenVertexArrays(1, &quadVAO);
        glGenBuffers(1, &quadVBO);
        glBindVertexArray(quadVAO);
        glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
        glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);
        glEnableVertexAttribArray(0);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindVertexArray(0);
    }
    glBindVertexArray(quadVAO);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glBindVertexArray(0);
}

// ʾϢ
void printInfos() {
    printf(
        "----------------------------------------------------------------------\n"
        "Use: SDF.exe [fragment shader] [BIG IMAGE]\n"
        "     BIG IMAGE: if has this tag, the resolution is changed to 1600*900\n"
        "\n"
        "Press W/A/S/D or Move Mouse to move the view point\n"
        "Press I/J, O/K, P/L to increase/decrease the light position (x, y, z)\n"
        "----------------------------------------------------------------------\n"
    );
}

// ÿһ֡ʾϢ
void printFrameInfo() {
    printf("Time:%6.2f, FPS:%8.2f\r", lastFrameTime, FPS);
}

// ¼
void processInput(GLFWwindow *window) {
    if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS)
        glfwSetWindowShouldClose(window, true);

    if (glfwGetKey(window, GLFW_KEY_W) == GLFW_PRESS)
        camera.ProcessKeyboard(FORWARD, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS)
        camera.ProcessKeyboard(BACKWARD, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS)
        camera.ProcessKeyboard(LEFT, deltaTime);
    if (glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS)
        camera.ProcessKeyboard(RIGHT, deltaTime);

    // Դ
    {
        const float delta = 0.01f;
        if (glfwGetKey(window, GLFW_KEY_I) == GLFW_PRESS)
            lightPosition.x += delta;
        if (glfwGetKey(window, GLFW_KEY_J) == GLFW_PRESS)
            lightPosition.x -= delta;
        if (glfwGetKey(window, GLFW_KEY_O) == GLFW_PRESS)
            lightPosition.y += delta;
        if (glfwGetKey(window, GLFW_KEY_K) == GLFW_PRESS)
            lightPosition.y -= delta;
        if (glfwGetKey(window, GLFW_KEY_P) == GLFW_PRESS)
            lightPosition.z += delta;
        if (glfwGetKey(window, GLFW_KEY_L) == GLFW_PRESS)
            lightPosition.z -= delta;
    }

    // ʾϢ
    {
        if (!QPressed && glfwGetKey(window, GLFW_KEY_Q) == GLFW_PRESS) {
            // camera
            camera.printCameraInfo();
            // light
            printf("Light Position: (%.2f, %.2f, %.2f)\n",
                lightPosition.x, lightPosition.y, lightPosition.z);
            QPressed = true;
        }
        if (glfwGetKey(window, GLFW_KEY_Q) == GLFW_RELEASE)
            QPressed = false;
    }
}

// ڱ仯¼
void framebufferSizeCallback(GLFWwindow* window, int width, int height) {
    glViewport(0, 0, width, height);
}

// ¼
void mouseCallback(GLFWwindow* window, double xpos, double ypos) {
    if (firstMouse) {
        lastX = (float)xpos;
        lastY = (float)ypos;
        firstMouse = false;
    }

    float xoffset = (float)xpos - lastX;
    // OpenGL  y 
    float yoffset = lastY - (float)ypos;

    lastX = (float)xpos;
    lastY = (float)ypos;

    camera.ProcessMouseMovement(xoffset, yoffset);
}

// ¼
void scrollCallback(GLFWwindow* window, double xoffset, double yoffset) {
    camera.ProcessMouseScroll((float)yoffset);
}

// 
void checkError() {
    int num = 0;
    GLenum a = GL_NO_ERROR;
    while (true) {
        a = glGetError();
        if (a == GL_NO_ERROR) {
            break;
        }
        printf("ERROR(%d): %d\n", num, a);
    }
}