#version 330 core

struct Camera {
    vec3 position;
    vec3 front;
    vec3 up;
    vec3 right;
    float focalLength; // ﲢûо۽˼, ֻǱʾƽ浽ľ
};

// 
uniform vec3    iResolution;  // viewport resolution (in pixels)
uniform int     iFrame;       // shader playback frame
uniform float   iTime;        // shader playback time (in seconds)

uniform Camera camera;
uniform vec3 lightPositionOffset;

#define LIGHT_POSITION vec3(-4.30, 5.45, 0.32)

// 
// : y=0
#define DESK_PLANE 1.0

// 
out vec4 FragColor;

#define AA 1
#define ZERO min(iFrame, 0)
#define PI 3.1415926

//-------------------------------------------------------------------------------
// 

float dot2( in vec3 v ) { return dot(v,v); }
float dot2( in vec2 v ) { return dot(v,v); }

// https://iquilezles.org/www/articles/sfrand/sfrand.htm
float sfrand( inout int mirand ) {
    mirand = mirand*0x343fd+0x269ec3;
    float res = uintBitsToFloat((uint(mirand)>>9)|0x40000000u );
    return( res-3.0 );
}

// hash by Hugo Elias(ϣ uint)
uint ihash( uint n ) {
    n = (n << 13) ^ n;
    n = (n*(n*n*15731u+789221u)+1376312589u)&0x7fffffffu;
    return n;
}

// hash by Hugo Elias(ϣ float)
float hash( int n ) {
    n = (n << 13) ^ n;
    n = (n*(n*n*15731+789221)+1376312589)&0x7fffffff;
    return 1.0 - float(n)*(1.0/1073741824.0);
}

float noise( vec3 p ) {
    ivec3 ip = ivec3(floor(p));
    vec3  fp = fract(p);

    vec3 w = fp*fp*(3.0-2.0*fp);

    int n = ip.x + ip.y*57 + 113*ip.z;

    return mix(mix(mix( hash(n+(0+57*0+113*0)),
                        hash(n+(1+57*0+113*0)),w.x),
                   mix( hash(n+(0+57*1+113*0)),
                        hash(n+(1+57*1+113*0)),w.x),w.y),
               mix(mix( hash(n+(0+57*0+113*1)),
                        hash(n+(1+57*0+113*1)),w.x),
                   mix( hash(n+(0+57*1+113*1)),
                        hash(n+(1+57*1+113*1)),w.x),w.y),w.z);
}

// https://iquilezles.org/www/articles/fbm/fbm.htm
// Fractional Brownian Motion(˶)
// 
float fbm( vec3 p ) {
#if 0
    // original code
    return 0.5000*noise( p*1.0 ) +
           0.2500*noise( p*2.0 ) +
           0.1250*noise( p*4.0 ) +
           0.0625*noise( p*8.0 );
#else
    // equivalent code, but compiles MUCH faster
    float f = 0.0;
    float s = 0.5;
    for( int i = ZERO; i < 4; i++ ) {
        f += s*noise( p );
        s *= 0.5;
        p *= 2.0;
    }
    return f;
#endif
}

// https://iquilezles.org/www/articles/smin/smin.htm
// polynomial smooth min 1 (k=0.1)
// ֮ĽСֵ, ƽ
float smin(float a, float b, float k){
    float h = clamp(0.5 + 0.5*(b - a)/k, 0.0, 1.0 );
    return mix(b, a, h) - k*h*(1.0 - h); // mix(x,y,a)=x*(1-a)+y*a
}

// ֮Ľϴֵ, ƽ
float smax(float a,float b, float k) {
    return -smin(-a, -b, k);
}

// ת
mat2 rotmat(float a) {
    float c = cos(a);
    float s = sin(a);
    // OpenGL 
    return mat2(c,s,-s,c);    // ҳ
    // return mat2(c,-s,s,c); // 
}


//-------------------------------------------------------------------------------
// ͼԪ
// http://iquilezles.org/www/articles/distfunctions/distfunctions.htm


// 
// p: Գĵ
// b:  x,y,z ķ(ʼΪ [-1,1]^3)
float sdBox(vec3 p, vec3 b) {
    // Ƶ: https://www.youtube.com/watch?v=62-pRVZuS5c&ab_channel=InigoQuilez
    //  3 ƽľ, abs 仯ɵһ
    vec3 d = abs(p) - b;
    // ߲: sdf<0(ڲ)
    // ұ߲: sdf>0(ⲿ)
    return min(max(d.x, max(d.y, d.z)), 0.0) + length(max(d, 0.0));
}

// ע udSqBox ص sdf^2
// , ͳ岻һĵطڲ sdf ֵΪ 0
float udSqBox( vec3 p, vec3 b ) {
    float di = 0.0;
    float dx = abs(p.x)-b.x; if( dx>0.0 ) di+=dx*dx;
    float dy = abs(p.y)-b.y; if( dy>0.0 ) di+=dy*dy;
    float dz = abs(p.z)-b.z; if( dz>0.0 ) di+=dz*dz;
    return di;
}

// Բ,  y=h,y=-h ƽ, ֻ֤ⲿȷ(sdf >=0)
// dir: (Ҫһ)
// h: ߶
// r: 뾶
float sdCylinder(vec3 p, vec3 dir, float h, float r) {
    float t = dot(p, dir);
    float d = distance(p, dir*t); // ֮ľ
    // ȡֵΪ
    return length( max( abs(vec2(d, t)) - vec2(r, h), vec2(0) ) );
}

// Ȧ(xoz)
// : (R-sqrt(x^2 + z^2))^2 + y^2 = r^2
// R: ĵܵĵľ
// r: ܵİ뾶
float sdTorus(vec3 p, float R, float r) {
    // ߵľȥܵİ뾶
    return length( vec2(length(p.xz)-R, p.y) )-r;
}

// еһεȦ(xoy)
// ra: ĵܵĵľ
// rb: ܵİ뾶
// sc: sin(x), cos(x), x ʾȦֵĴС x=[0, 3.14]
float sdCappedTorus(in vec3 p, in vec2 sc, in float ra, in float rb) {
    p.x = abs(p.x); // Գ,  1,4 
    // һ, β㵽Բ(cap)ľ, 㵽Բ()ľ
    // · ? case1 : case2
    float k = (sc.y*p.x>sc.x*p.y) ? dot(p.xy, sc) : length(p.xy);
    // Ϸ, ʽչΪʽ
    // ·,  cap ĵľ, չ
    return sqrt( dot(p,p) + ra*ra - 2.0*ra*k ) - rb;
}


//-------------------------------------------------------------------------------

// 
float sdDesk( vec3 pos ) {
    float dist = 1e10;

    // 
    dist = min(dist, sdBox( pos - vec3(0.0, DESK_PLANE-0.01, 0.0), vec3(1.0, 0.02, 1.0) ));
    // 
    const float[] dx = float[4]( 1.0, -1.0, -1.0,  1.0 );
    const float[] dy = float[4]( 1.0,  1.0, -1.0, -1.0 );
    for( int i = ZERO; i < 4; ++i ) {
        dist = min(dist, sdBox(
            pos - vec3(dx[i]*0.8, DESK_PLANE/2.0, dy[i]*0.8),
            vec3(0.02, DESK_PLANE/2.0, 0.02) ));
    };
    return dist;
}

// 鱾
float sdBook(vec3 pos, out uint subMatID) {
    float dist = 1e10;
    subMatID = 22u;
    pos -= vec3(-0.5, DESK_PLANE, -0.6);

    dist = min(dist, sdBox(pos - vec3(0.0, 0.06, 0.0), vec3(0.15, 0.04, 0.2)));   // b1 ҳ
    {
        float d = 1e10;
        d = min(d, sdBox(pos - vec3(0.01, 0.01, 0.0), vec3(0.16, 0.01, 0.21)));   // b1 
        d = min(d, sdBox(pos - vec3(0.01, 0.11, 0.0), vec3(0.16, 0.01, 0.21)));   // b1 
        d = min(d, sdBox(pos - vec3(-0.15, 0.06, 0.0), vec3(0.01, 0.06, 0.21)));  // b1 
        if(d < dist) {
            dist = d; subMatID = 2u;
        }
    }

    pos.xz *= rotmat(0.3);
    {
        float d = min(dist, sdBox(pos - vec3(0.0, 0.18, 0.0), vec3(0.12, 0.04, 0.15)));  // b2 ҳ
        if(d < dist) {
            dist = d; subMatID = 22u;
        }
    }
    {
        float d = 1e10;
        d = min(d, sdBox(pos - vec3(0.01, 0.13, 0.0), vec3(0.13, 0.01, 0.16)));   // b2 
        d = min(d, sdBox(pos - vec3(0.01, 0.23, 0.0), vec3(0.13, 0.01, 0.16)));   // b2 
        d = min(d, sdBox(pos - vec3(-0.12, 0.18, 0.0), vec3(0.01, 0.06, 0.16)));  // b2 
        if(d < dist) {
            dist = d; subMatID = 3u;
        }
    }


    return dist;
}

// 
float sdCup(vec3 pos) {

    //  21u
#define CUP_r 0.01
#define CUP_R 0.07
#define CUP_HEIGHT 0.2
    pos -= vec3(0.7, DESK_PLANE+2.0*CUP_r, -0.8);
    float dist = sdTorus( pos-vec3(0.0,clamp(pos.y,0.0,CUP_HEIGHT),0.0), CUP_R, CUP_r);

    // Ե 21u
    dist = smin(dist, sdTorus(pos-vec3(0.0,CUP_HEIGHT,0.0), CUP_R, CUP_r), 0.01);

    //  21u
    dist = smin(dist, sdCylinder(pos, vec3(0.0,1.0,0.0), CUP_r, CUP_R), 0.05);

    // ӵİ 21u
    vec2 sc = vec2(1.0, 0.0); // Ȧ
    vec3 p = pos-vec3(CUP_R,CUP_HEIGHT/2.0,0.0);
    p.xy *= mat2(0.0,-1.0,1.0,0.0); // ˳ʱת 90 
    dist = smin(dist, sdCappedTorus(p, sc, CUP_HEIGHT/4.0, CUP_r), 0.01);

    return dist;
}

// ߵĳ
float sdMonster(vec3 pos) {
    float dist = 1e10;
    // (1+abs(cos(1.8)))*0.2+0.07 = 0.31544( R,r)
    #define BODY_CENTER 0.31544
    pos -= vec3(0.0, BODY_CENTER+DESK_PLANE, 0.0);
    float iTimesin = sin(iTime*2.0);

    // 
    {
        // Ȧ
        vec3 p = pos;
        p.xz *= rotmat(0.4*iTimesin); // 
        vec2 sc = vec2(sin(1.8), cos(1.8));
        p.y += 0.2; // R=0.2
        dist = sdCappedTorus(p, sc, 0.2, 0.07); // R=0.2,r=0.07
    }

    // 
    {
        // Բ
        dist = smin(dist, distance( vec3(0.0, clamp(pos.y, 0.0, 0.2), 0.0) , pos) - 0.07, 0.2);
    }


    pos.y -= 0.2; // ߶Ϊ 0.2
    // 
    {
        // Ȧ
        vec3 p = pos;
        p.xz *= rotmat(-0.5*iTimesin); // 
        vec2 sc = vec2(sin(0.45), cos(0.45));
        p.y -= 0.5; // R=0.5, r=0.07
        p.y = -p.y;
        dist = smin(dist, sdCappedTorus(p, sc, 0.5, 0.07), 0.2);
    }

    pos.y -= 0.07; // r=0.07
    // ͷ
    {
        // 
        vec3 p = pos;
        p.y -= 0.1 + (iTimesin+1.0)*0.02;
        dist = min(dist, distance(p, vec3(0.0,  clamp(p.y,0.0,0.02), 0.0)) - 0.1);
    }

    return dist;
}

float sdApple(vec3 pos, out uint subMatID) {
    float dist = 1e10;
    pos -= vec3(-0.6,DESK_PLANE,0.1);

    #define APPLE_R 0.1
    #define ALLPE_DOWN_OFF 0.04
    #define ALLPE_UP_OFF 0.03
    // ƻ
    subMatID = 24u;
    dist =      sdTorus( pos-vec3(0.0, 2.0*APPLE_R-ALLPE_DOWN_OFF, 0.0), APPLE_R-ALLPE_UP_OFF, APPLE_R);
    dist = smin(sdTorus( pos-vec3(0.0,   APPLE_R-ALLPE_DOWN_OFF, 0.0), APPLE_R-ALLPE_UP_OFF, APPLE_R-ALLPE_DOWN_OFF), dist, 0.1);

    // ƻ
    pos.y -= (2.0*APPLE_R-ALLPE_DOWN_OFF);
    {
        #define APPLE_H 0.08
        pos.y -= APPLE_H;
        pos.xy *= rotmat(3.0*pos.y + noise(pos));
        float d = sdCylinder(pos, vec3(0.0,1.0,0.0), APPLE_H, 0.005+0.02*pos.y);
        if(d < dist) {
            dist = d;
            subMatID = 4u;
        }
    }
    return dist;
}

// 
float map( vec3 pos, out uint sid, out uint subMat ) {
    subMat = 0u;           // еĲ
    float mindist = 1e10;  // sdf ֵ

    //-----------------------
    // 
    // sid=2u
    //-----------------------
    {
        float dist = sdDesk(pos);
        if(dist < mindist) {
            sid = 2u; subMat = 1u;
            mindist = dist;
        };
    }

#if 1
    //-----------------------
    // suelo ذ
    // sid=1u
    //-----------------------
    {
#define suelo_y (0.0)
        float dis = pos.y - suelo_y;

        // ƵשĴС, suelo_seg ԽԽС
#define suelo_seg 2.0
        float ax = 128.0 + (pos.x+pos.z)*suelo_seg;
        float az = 128.0 + (pos.x-pos.z)*suelo_seg;
        uint ix = uint(floor(ax));
        uint iz = uint(floor(az));
        uint kk = ihash(ix+53u*iz); // ھذϵĸӵɫ

        bool ba = (((subMat>>10)&7u) > 6u);
        float peld = max(fract(ax),fract(az)); // fract(x) = x - floor(x)
        // smoothstep(edge0, edge1, t)
        // ȼ
        //     t = clamp((x - edge0) / (edge1 - edge0), 0.0, 1.0);
        //     t = (3.0 - 2.0 * t) * t * t;
        peld = smoothstep(0.975, 1.0, peld); //  peld ֵ [0.975, 1.0] ֮
        if(ba) peld = 1.0;
        dis += 0.005*peld;        // ɵߵͲƽĸо(ÿһשһ)
        if(dis < mindist) {
            mindist = dis;
            subMat = kk;
            // sid=0u(ɫ), sid=1u(ذ)
            if( peld > 0.0000001 ) sid = 0u; else sid = 1u;
        }
    }

    //-----------------------
    // ϵİ
    // sid=2u
    //-----------------------

    // 
    {
        float dist = sdCup(pos);
        if(dist < mindist) {
            sid = 2u; subMat = 21u;
            mindist = dist;
        };
    }

    // 鱾
    {
        uint kk;
        float dist = sdBook(pos, kk);
        if(dist < mindist) {
            sid = 2u; subMat = kk;
            mindist = dist;
        };
    }

    // ߵĳ
    {
        float dist = sdMonster(pos);
        if(dist < mindist) {
            sid = 2u; subMat = 23u;
            mindist = dist;
        };
    }

#endif
    // ƻ
    {
        uint kk;
        float dist = sdApple(pos, kk);
        if(dist < mindist) {
            sid = 2u; subMat = kk;
            mindist = dist;
        };
    }

    return mindist;
}

void raycast( vec3 ro, vec3 rd, float tMin, float tMax,
              out float resT, out vec3 resPos,
              out uint resSurface, out uint subMat ) {
    float t = tMin;
    for(int i = ZERO; i < 256 && t < tMax; i++) {
        resPos = ro + t*rd;
        float h = map(resPos, resSurface, subMat);
        if(h < 0.0001) {
            resT = t;
            return;
        }
        // 򵥵Ĳʽ, 0.75 һ
        t += h*0.75;
    }
    // ûκ resSurface=666u
    resSurface = 666u;
}

vec3 calcNormal( vec3 pos ) {
#if 0
    // original code
    const float eps = 0.0002;
    uint kk, kk2;
    vec3 nor = vec3(
      map( pos.x+eps, pos.y, pos.z, kk, kk2 ) - map( pos.x-eps, pos.y, pos.z, kk, kk2 ),
      map( pos.x, pos.y+eps, pos.z, kk, kk2 ) - map( pos.x, pos.y-eps, pos.z, kk, kk2 ),
      map( pos.x, pos.y, pos.z+eps, kk, kk2 ) - map( pos.x, pos.y, pos.z-eps, kk, kk2 ));
    return normalize(nor);
#else
    // inspired by tdhooper and klems - a way to prevent the compiler from inlining map() 4 times
    vec3 n = vec3(0.0);
    for(int i = ZERO; i < 4; i++) {
        uint kk, kk2;
        vec3 e = 0.5773*(2.0*vec3((((i+3)>>1)&1),((i>>1)&1),(i&1))-1.0);
        n += e*map(pos + 0.0002*e, kk, kk2);
        if( n.x + n.y + n.z>100.0 ) break;
    }
    return normalize(n);
#endif
}

// http://iquilezles.org/www/articles/rmshadows/rmshadows.htm
float calcSoftshadow( in vec3 ro, in vec3 rd, in float mint, in float tmax ) {
    float res = 1.0;
    float t = mint;
    for(int i = ZERO; i < 128; i++) {
        uint tt1, tt2;
        float h = map(ro + rd*t, tt1, tt2);
        // 򵥵 softshadow(ûиĽ)
        // k=8.0
        float s = clamp(8.0*h/t, 0.0, 1.0);
        res = min( res, s );
        // t += clamp( h, 0.02, 0.2 );
        t += h;
        if( t > tmax ) break;
    }
    res = res*res*(3.0 - 2.0*res); // 0.037 => 0.004
    return clamp( res, 0.0, 1.0 );
}

float calcAO( in vec3 pos, in vec3 nor ) {
    float occ = 0.0;
    float sca = 1.0;
    for( int i = ZERO; i < 5; i++ ) {
        float h = 0.01 + 0.12*float(i)/4.0;
        uint tt1, tt2;
        float d = map(pos + h*nor, tt1, tt2);
        occ += (h - d)*sca;
        sca *= 0.95;
        if(occ > 0.35) break;
    }
    return clamp(1.0 - 3.0*occ, 0.0, 1.0) * (0.5 + 0.5 * nor.y);
}

// 
// subMatID ΪҪ bump mapping
vec3 shade( vec3 pos, vec3 rd, uint matID, uint subMatID, float dis ) {
    // ûлκ
    if( matID == 666u ) return vec3(0.0,0.0,0.0);

    vec3 col = vec3(0.1 * dis);

    // calc normal
    vec3 nor = calcNormal( pos );

    // bump mapping
    // Է߽Ŷ
#if 1
    if(matID != 2u || subMatID <= 20u) {
        const float kke = 0.0001;
        float bumpa = 0.0075; // Խ͹̶Խ
        bumpa *= 0.75;
        bumpa /= kke;
        float kk = fbm( 32.0*pos );
        nor.x += bumpa*(fbm( 32.0*vec3(pos.x+kke, pos.y, pos.z) )-kk);
        nor.y += bumpa*(fbm( 32.0*vec3(pos.x, pos.y+kke, pos.z) )-kk);
        nor.z += bumpa*(fbm( 32.0*vec3(pos.x, pos.y, pos.z+kke) )-kk);
        nor = normalize(nor);
    }
#endif

    // materials
    if( matID == 0u ) {
        // ID=0, ȫɫ
        col = vec3(0.0, 0.0, 0.0);
    }
    else if( matID == 1u ) {
        // ID=1 ذ
        // subMatID Ϊһλص
        // ͨһļвͬЧ
        float xoff = 13.1 * float(subMatID&255u);
        float fb = fbm( 16.0*vec3(pos.x+xoff, pos.y, pos.z) );

        col = 0.7 + vec3(0.20, 0.22, 0.25) * fb;

        float baldscale = float((subMatID>>9)&15u)/14.0;
        baldscale = 0.51 + 0.34*baldscale;
        col *= baldscale;
        float fx = 1.0;
        if( (subMatID&256u)!=0u ) fx = -1.0;
        float m = sin( 64.0*pos.z*fx +  64.0*pos.x + 4.0*fb );
        m = smoothstep( 0.25, 0.5, m ) - smoothstep( 0.5, 0.75, m );

        col += m*0.15;
        col *= vec3(0.5,0.5,0.2); // ɫ
    }
    else if( matID == 2u ) {
        // bump mapping
        // 1u 
        if( subMatID == 1u ) col = vec3(0.5);
        // 2u  1 
        else if( subMatID == 2u ) col = vec3(0.2, 0.05, 0.05);
        // 3u  2 
        else if( subMatID == 3u ) col = vec3(0.05, 0.2, 0.05);
        // 4u ƻ
        else if( subMatID == 4u ) col = vec3(0.247, 0.112, 0.112);

        // no bump mapping
        // 21u 
        else if( subMatID == 21u ) col = vec3(0.6,0.1,0.1);
        // 22u 鱾ڵİɫ鱾
        else if( subMatID == 22u ) col = vec3(1.0);
        // 23u ߵĳ
        else if( subMatID == 23u ) col = vec3(0.05, 0.2, 0.5);
        // 24u ƻ
        else if( subMatID == 24u ) col = vec3(0.5, 0.0, 0.0);
    }

    // shade
    // blinn-phong
    vec3  lig = normalize( LIGHT_POSITION + lightPositionOffset - pos );
    vec3  hal = normalize( lig - rd );
    float dif = clamp( dot( nor, lig ), 0.0, 1.0 );
    float spe = pow( clamp( dot( nor, hal ), 0.0, 1.0 ), 16.0);

    float ao = calcAO(pos, nor); // AO
    float ss = calcSoftshadow( pos, lig, 0.02, 20.0 ); // shadow
    col *= (ao*0.05 + ss*(dif + spe*vec3(0.2)));

    return col;
}

// 
vec3 post( vec3 col) {

    // gamma correction
    col = pow( col, vec3(0.4545) ); // 1/2.2=0.4545

    return clamp(col, 0.0, 1.0);
}

vec3 computePixel(vec2 p) {
    // y: [-1, 1]
    // x: [-a, a], a=iResolution.x/iResolution.y
    vec2 s = (2.0*p - iResolution.xy)/iResolution.y;

    // camera, ray
    mat3 ca = mat3(camera.right, camera.up, camera.front);
    vec3 rd = ca * normalize(vec3(s, camera.focalLength));

    // 
    float t;              // ʱ
    uint matID, sumMatID; // 
    vec3 xyz;             // λ

    // raycast
    // tMin ó 0.5 Ϊ˱תʱڵס
    raycast( camera.position, rd, 0.0, 24.0, t, xyz, matID, sumMatID );

    // shade
    vec3 col = shade( xyz, rd, matID, sumMatID, t );

    // Դ
    {
        vec3 v1 = lightPositionOffset + LIGHT_POSITION - camera.position;
        float lig_dist2 = dot(v1, v1);
        float dist = dot(v1 ,rd);
        float seg2 = lig_dist2 - dist*dist;
#if 1
        // ״Դ
        // (1) Դǰ / ûл
        if((dist > 0.0 && (dist < t || matID == 666u))) {
            col += vec3(1.0, 0.95, 0.90)*exp2( -64.0*seg2 );
        }
#else
        // йԴֱӷ
        #define R2 0.01
        // (1) Դǰ / ûл
        // (2) йԴ
        if((dist > 0 && (dist < t || matID == 666u)) && (seg2 < R2)) {
            return vec3( 1.0, 0.95, 0.90 );
        }
#endif
    }

    // postprocess
    col = post(col);

    return col;
}

void main() {
#if AA<2
    vec3 col = computePixel(gl_FragCoord.xy);
#else
    vec3 col = vec3(0.0);
    for( int j = 0; j < AA; j++ ) {
        for( int i = 0; i < AA; i++ ) {
            vec2 of = vec2(i, j) / float(AA);
            col += computePixel(gl_FragCoord.xy + of);
        }
        col /= float(AA*AA);
    }
#endif
    FragColor = vec4(col, 1.0);
}