// shader toy: https://www.shadertoy.com/view/XlccWH
#version 330 core

struct Camera {
    vec3 position;
    vec3 front;
    vec3 up;
    vec3 right;
    float focalLength; // ﲢûо۽˼, ֻǱʾƽ浽ľ
};

// 
uniform vec3    iResolution;  // viewport resolution (in pixels)
uniform int     iFrame;       // shader playback frame
uniform float   iTime;        // shader playback time (in seconds)

uniform Camera camera;
uniform vec3 lightPositionOffset;

#define LIGHT_POSITION vec3(-4.30, 5.45, -6.54)

// 
out vec4 FragColor;

#define AA 1
#define ZERO min(iFrame, 0)
#define PI 3.1415926

//-------------------------------------------------------------------------------
// 

float dot2( in vec3 v ) { return dot(v,v); }
float dot2( in vec2 v ) { return dot(v,v); }

// https://iquilezles.org/www/articles/sfrand/sfrand.htm
float sfrand( inout int mirand ) {
    mirand = mirand*0x343fd+0x269ec3;
    float res = uintBitsToFloat((uint(mirand)>>9)|0x40000000u );
    return( res-3.0 );
}

// hash by Hugo Elias(ϣ uint)
uint ihash( uint n ) {
    n = (n << 13) ^ n;
    n = (n*(n*n*15731u+789221u)+1376312589u)&0x7fffffffu;
    return n;
}

// hash by Hugo Elias(ϣ float)
float hash( int n ) {
    n = (n << 13) ^ n;
    n = (n*(n*n*15731+789221)+1376312589)&0x7fffffff;
    return 1.0 - float(n)*(1.0/1073741824.0);
}

float noise( vec3 p ) {
    ivec3 ip = ivec3(floor(p));
    vec3  fp = fract(p);

    vec3 w = fp*fp*(3.0-2.0*fp);

    int n = ip.x + ip.y*57 + 113*ip.z;

    return mix(mix(mix( hash(n+(0+57*0+113*0)),
                        hash(n+(1+57*0+113*0)),w.x),
                   mix( hash(n+(0+57*1+113*0)),
                        hash(n+(1+57*1+113*0)),w.x),w.y),
               mix(mix( hash(n+(0+57*0+113*1)),
                        hash(n+(1+57*0+113*1)),w.x),
                   mix( hash(n+(0+57*1+113*1)),
                        hash(n+(1+57*1+113*1)),w.x),w.y),w.z);
}

// https://iquilezles.org/www/articles/fbm/fbm.htm
// Fractional Brownian Motion(˶)
// 
float fbm( vec3 p ) {
#if 0
    // original code
    return 0.5000*noise( p*1.0 ) +
           0.2500*noise( p*2.0 ) +
           0.1250*noise( p*4.0 ) +
           0.0625*noise( p*8.0 );
#else
    // equivalent code, but compiles MUCH faster
    float f = 0.0;
    float s = 0.5;
    for( int i = ZERO; i < 4; i++ ) {
        f += s*noise( p );
        s *= 0.5;
        p *= 2.0;
    }
    return f;
#endif
}

// https://iquilezles.org/www/articles/smin/smin.htm
// polynomial smooth min 1 (k=0.1)
// ֮ĽСֵ, ƽ
float smin(float a, float b, float k){
    float h = clamp(0.5 + 0.5*(b - a)/k, 0.0, 1.0 );
    return mix(b, a, h) - k*h*(1.0 - h); // mix(x,y,a)=x*(1-a)+y*a
}

// ֮Ľϴֵ, ƽ
float smax(float a,float b, float k) {
    return -smin(-a, -b, k);
}

// ת
mat2 rotmat(float a) {
    float c = cos(a);
    float s = sin(a);
    // OpenGL 
    return mat2(c,s,-s,c);    // ҳ
    // return mat2(c,-s,s,c); // 
}


//-------------------------------------------------------------------------------
// ͼԪ

// http://iquilezles.org/www/articles/distfunctions/distfunctions.htm
float sdBox( in vec3 p, in vec3 b ) {
    vec3 q = abs(p) - b;
    return min(max(q.x,max(q.y,q.z)),0.0) + length(max(q,0.0));
}

// http://iquilezles.org/www/articles/distfunctions/distfunctions.htm
// ע udSqBox ص sdf^2
// , ͳ岻һĵطڲ sdf ֵΪ 0
float udSqBox( vec3 p, vec3 b ) {
    float di = 0.0;
    float dx = abs(p.x)-b.x; if( dx>0.0 ) di+=dx*dx;
    float dy = abs(p.y)-b.y; if( dy>0.0 ) di+=dy*dy;
    float dz = abs(p.z)-b.z; if( dz>0.0 ) di+=dz*dz;
    return di;
}

// Բ,  y=h,y=-h ƽ, ֻ֤ⲿȷ(sdf >=0)
// dir: (Ҫһ)
// h: ߶
// r: 뾶
float cylinder(vec3 p, vec3 dir, float h, float r) {
    float t = dot(p, dir);
    float d = distance(p, dir*t); // ֮ľ
    // ȡֵΪ
    return length( max( abs(vec2(d, t)) - vec2(r, h), vec2(0) ) );
}

//-------------------------------------------------------------------------------
float pupdist = 1e4;
float rabdist(vec3 p) {

    float time2 = iTime * 2.4;

    // ζЧ
    p.y += 2.3;
    p.xy *= rotmat(cos(time2+1.0) * 0.04);
    p.y -= 2.3;

    vec3 op = p; // origin p

    // p2: ͷ(ͷһζЧ)
    vec3 p2 = p;
    p2.xy *= rotmat(cos(time2)*0.1);

    // p3: 
    vec3 p3 = p;
    p3.xy *= rotmat(cos(time2-length(p)/2.0)*0.13);

    float d = 1e4;
    // ȡֵ, ҶԳ
    p.x = abs(p.x);
    p2.x = abs(p2.x);
    p3.x = abs(p3.x);

    // Head
    // : ں
    d = smin(length(p2 - vec3(-0.75, 0.0, -0.1)) - 0.4,
             length(p2 - vec3( 0.75, 0.0, -0.1)) - 0.5, 2.0);
    d = smin(d,
             length(p2 - vec3(  0.0, 0.4, -0.1)) - 0.9, 1.6);
    d += 0.1;

    // Ears 1
    // Բ: Բ, 뾶 0.4
    d = smin(d, distance( vec3(0.7, clamp(p3.y, 0.0, 2.2), 0.0) ,p3 ) - 0.4, 0.14);
    // : (γɶϵİ)
    d = smax(d, -(length( p3 - vec3(0.7, 1.7, -0.5) ) - 0.5), 0.2);

    // Ears 2
    // : ϵСͻ
    d = smin(d, distance(vec3(1.1, 2.3, -0.1), p3) - 0.2, 0.8);

    // Neck
    // Բ
    d = smin(d, distance(vec3(0.0, clamp(p.y, -1.6, -1.1), 0.0), p.xyz) - 0.6, 0.04);

    // Legs
    float iTime2 = iTime * 2.0;
    vec3 p4 = op; // ҽжһ
    // yoz: ת(ld, 0)
    float ld = -0.75;
    p4.y -= ld;
    p4.yz *= rotmat(cos(iTime2) * 0.1);
    p4.y += ld;
    // Բ
    p4.y -= max(0.0, cos(iTime2 + PI / 2.0)) * 0.1;
    d = smin(d, distance(vec3(.3,clamp(p4.y,-2.6,-2.),0.),p4.xyz)-.3,.1);
    // һֻ
    p4 = op;
    p4.y -= ld;
    p4.yz *= rotmat(cos(iTime2 + PI) * 0.1);
    p4.y += ld;
    p4.y -= max(0.0, cos(iTime2 + PI / 2.0 + PI)) * 0.1;
    d = smin(d, distance( vec3(0.3, clamp(p4.y, -2.6, -2.0), 0.0), p4 * vec3(-1 ,1 ,1) ) -0.3, 0.1);

    // Belly
    // : 
    d = smin(d, distance(vec3(0.0, -1.5, -0.2), p) - 0.5 + cos(iTime*3.0)*0.03, 0.4);

    // Tail
    // 
    d = smin(d, distance(vec3(0, -1.7, 0.6), p) - 0.3, 0.1);


    // Eyes
    vec3 q = vec3(0.35, 0.4, -1.0);
    if(mod(iTime - 1.0, 4.0) > 0.04) {
        // <0.04, 1% ĸʲɼ(գ)
        // Բ, ۿ, Բ
        d = smax(d, -(cylinder(p2 - q, normalize(q - p2), 0.3, 0.1) - 0.0001), 0.05);
        // 
        d = smin(d, (length(p2 - q*0.9)- 0.2), 0.24);

        // Eye pupils
        // : Բ
        if(op.x > 0.0)
            pupdist = (length(p2 - vec3(0.39, 0.32,  -1.0)) - 0.2);
        else
            pupdist = (length(p2 - vec3(0.28, 0.32, -1.02)) - 0.2);

        d = smin(d, pupdist, 0.005);
    }

    // Nose
    // 
    d = smin(d, (length(p2 - vec3(0, 0.1, -1.02)) - 0.2), 0.02);

    // Mouth (two states)
    // Բ: Բ smax
    float d3 = smax(-(length(p-vec3(-.05,-.29,-1.02))-.1),-(length(p-vec3(.05,-.29,-1.02))-.1),.1);
    // 
    float d2 = max(p2.z, distance( p2,
                    vec3(clamp(p2.x, 0.0, 0.3), -0.2, clamp(p2.z, -2.0, 2.0))) + 0.01 );
    float time4 = iTime / 8.0;
    float gg = smoothstep(0.0 , 1.0,
                   clamp( (min(fract(time4), 1.0 - fract(time4)) - 0.25)*64.0 , 0.0, 1.0 ));
    d = smax(d, mix(-d2, d3, gg), 0.1);

    // Tooth
    // 
    d = min(d, (length(p - vec3(0.0, -0.2, -1.02)) - 0.08));


    // Arms
    // չ
    p.xy *= rotmat(0.4 + cos(time2*2.0) * 0.02);
    // ƽת
    p.y += 0.2;
    // +Բ
    float armd = smin(distance(vec3(0.2, clamp(p.y, -1.8, -0.0), 0.0), p) - 0.2,
                      distance(vec3(0.2, -1.7, 0.0) , p) - 0.2,
                      0.2);

    d = smin(d, armd, 0.05);
    return d;
}

// 
float map( vec3 pos, out uint sid, out uint subMat ) {
    subMat = 0u;           // еĲ
    float mindist = 1e10;  // sdf ֵ

    //-----------------------
    // suelo ذ
    //-----------------------
    {
#define suelo_y (-2.85)
        float dis = pos.y - suelo_y;

        // ƵשĴС, suelo_seg ԽԽС
#define suelo_seg 2.0
        float ax = 128.0 + (pos.x+pos.z)*suelo_seg;
        float az = 128.0 + (pos.x-pos.z)*suelo_seg;
        uint ix = uint(floor(ax));
        uint iz = uint(floor(az));
        subMat = ihash(ix+53u*iz); // ھذϵĸӵɫ

        bool ba = (((subMat>>10)&7u) > 6u);
        float peld = max(fract(ax),fract(az)); // fract(x) = x - floor(x)
        // smoothstep(edge0, edge1, t)
        // ȼ
        //     t = clamp((x - edge0) / (edge1 - edge0), 0.0, 1.0);
        //     t = (3.0 - 2.0 * t) * t * t;
        peld = smoothstep(0.975, 1.0, peld); //  peld ֵ [0.975, 1.0] ֮
        if(ba) peld = 1.0;
        dis += 0.005*peld;        // ɵߵͲƽĸо(ÿһשһ)
        mindist = dis;
        // sid=0u(ɫ), sid=1u(ذ)
        if( peld > 0.0000001 ) sid = 0u; else sid = 1u;
    }

    // Բ
    {
        float dist = rabdist(pos);
        if( dist < mindist ) {
            mindist = dist;
            sid = 2u;
        }
    }
    return mindist;
}

void raycast( vec3 ro, vec3 rd, float tMin, float tMax,
              out float resT, out vec3 resPos,
              out uint resSurface, out uint subMat ) {
    float t = tMin;
    for(int i = ZERO; i < 256; i++) {
        resPos = ro + t*rd;
        float h = map(resPos, resSurface, subMat);
        if(h < 0.0001) {
            resT = t;
            return;
        }
        // 򵥵Ĳʽ, 0.75 һ
        t += h*0.75;
        if(t > tMax) break;
    }
    // ûκ resSurface=666u
    resSurface = 666u;
}

vec3 calcNormal( vec3 pos ) {
#if 0
    // original code
    const float eps = 0.0002;
    uint kk, kk2;
    vec3 nor = vec3(
      map( pos.x+eps, pos.y, pos.z, kk, kk2 ) - map( pos.x-eps, pos.y, pos.z, kk, kk2 ),
      map( pos.x, pos.y+eps, pos.z, kk, kk2 ) - map( pos.x, pos.y-eps, pos.z, kk, kk2 ),
      map( pos.x, pos.y, pos.z+eps, kk, kk2 ) - map( pos.x, pos.y, pos.z-eps, kk, kk2 ));
    return normalize(nor);
#else
    // inspired by tdhooper and klems - a way to prevent the compiler from inlining map() 4 times
    vec3 n = vec3(0.0);
    for(int i = ZERO; i < 4; i++) {
        uint kk, kk2;
        vec3 e = 0.5773*(2.0*vec3((((i+3)>>1)&1),((i>>1)&1),(i&1))-1.0);
        n += e*map(pos + 0.0002*e, kk, kk2);
        if( n.x + n.y + n.z>100.0 ) break;
    }
    return normalize(n);
#endif
}

// http://iquilezles.org/www/articles/rmshadows/rmshadows.htm
float calcSoftshadow( in vec3 ro, in vec3 rd, in float mint, in float tmax ) {
    float res = 1.0;
    float t = mint;
    for(int i = ZERO; i < 24; i++) {
        uint tt1, tt2;
        float h = map(ro + rd*t, tt1, tt2);
        // 򵥵 softshadow(ûиĽ)
        // k=8.0
        float s = clamp(8.0*h/t, 0.0, 1.0);
        res = min( res, s );
        // t += clamp( h, 0.02, 0.2 );
        t += h;
        if( t > tmax ) break;
    }
    res = res*res*(3.0 - 2.0*res); // 0.037 => 0.004
    return clamp( res, 0.0, 1.0 );
}

float calcAO( in vec3 pos, in vec3 nor ) {
    float occ = 0.0;
    float sca = 1.0;
    for( int i = ZERO; i < 5; i++ ) {
        float h = 0.01 + 0.12*float(i)/4.0;
        uint tt1, tt2;
        float d = map(pos + h*nor, tt1, tt2);
        occ += (h - d)*sca;
        sca *= 0.95;
        if(occ > 0.35) break;
    }
    return clamp(1.0 - 3.0*occ, 0.0, 1.0) * (0.5 + 0.5 * nor.y);
}

// 
vec3 shade( vec3 pos, vec3 rd, uint matID, uint subMatID, float dis ) {
    // ûлκ
    if( matID == 666u ) return vec3(0.0,0.0,0.0);

    vec3 col = vec3(0.1 * dis);

    // calc normal
    vec3 nor = calcNormal( pos );

    // bump mapping
    // Է߽Ŷ
#if 1
    if(matID != 2u) {
        const float kke = 0.0001;
        float bumpa = 0.0075; // Խ͹̶Խ
        bumpa *= 0.75;
        bumpa /= kke;
        float kk = fbm( 32.0*pos );
        nor.x += bumpa*(fbm( 32.0*vec3(pos.x+kke, pos.y, pos.z) )-kk);
        nor.y += bumpa*(fbm( 32.0*vec3(pos.x, pos.y+kke, pos.z) )-kk);
        nor.z += bumpa*(fbm( 32.0*vec3(pos.x, pos.y, pos.z+kke) )-kk);
        nor = normalize(nor);
    }
#endif

    // materials
    if( matID == 0u ) {
        // ID=0, ȫɫ
        col = vec3(0.0, 0.0, 0.0);
    }
    else if( matID == 1u ) {
        // ID=1 ذ
        // subMatID Ϊһλص
        // ͨһļвͬЧ
        float xoff = 13.1 * float(subMatID&255u);
        float fb = fbm( 16.0*vec3(pos.x+xoff, pos.y, pos.z) );

        col = 0.7 + vec3(0.20, 0.22, 0.25) * fb;

        float baldscale = float((subMatID>>9)&15u)/14.0;
        baldscale = 0.51 + 0.34*baldscale;
        col *= baldscale;
        float fx = 1.0;
        if( (subMatID&256u)!=0u ) fx = -1.0;
        float m = sin( 64.0*pos.z*fx +  64.0*pos.x + 4.0*fb );
        m = smoothstep( 0.25, 0.5, m ) - smoothstep( 0.5, 0.75, m );

        col += m*0.15;
        col *= vec3(0.5,0.5,0.2); // ɫ
    }
    else if( matID == 2u ) {
        float gray = mix(0.1, 0.4, smoothstep(0.0, .01, pupdist));
        if(gray <= 0.11) {
            col = vec3(0.0);
        } else {
            col = vec3(0.8,0.1,0.1);
            // col = vec3(0.8);
        }
    }

    // shade
    // blinn-phong
    vec3  lig = normalize( LIGHT_POSITION + lightPositionOffset - pos );
    vec3  hal = normalize( lig - rd );
    float dif = clamp( dot( nor, lig ), 0.0, 1.0 );
    float spe = pow( clamp( dot( nor, hal ), 0.0, 1.0 ), 16.0);

    float ao = calcAO(pos, nor); // AO
    float ss = calcSoftshadow( pos, lig, 0.02, 20 ); // shadow
    col *= (ao*0.02*dif + ss*(dif + spe*vec3(0.2)));

    return col;
}

// 
vec3 post( vec3 col) {

    // gamma correction
    col = pow( col, vec3(0.4545) ); // 1/2.2=0.4545

    return clamp(col, 0.0, 1.0);
}

vec3 computePixel(vec2 p) {
    // y: [-1, 1]
    // x: [-a, a], a=iResolution.x/iResolution.y
    vec2 s = (2.0*p - iResolution.xy)/iResolution.y;

    // camera, ray
    mat3 ca = mat3(camera.right, camera.up, camera.front);
    vec3 rd = normalize(ca * vec3(s, camera.focalLength));

    // 
    float t;              // ʱ
    uint matID, sumMatID; // 
    vec3 xyz;             // λ

    // raycast
    // tMin ó 0.5 Ϊ˱תʱڵס
    raycast( camera.position, rd, 0, 24.0, t, xyz, matID, sumMatID );

    // shade
    vec3 col = shade( xyz, rd, matID, sumMatID, t );

    // Դ
    {
        vec3 v1 = lightPositionOffset + LIGHT_POSITION - camera.position;
        float lig_dist2 = dot(v1, v1);
        float dist = dot(v1 ,rd);
        float seg2 = lig_dist2 - dist*dist;
#if 1
        // ״Դ
        // (1) Դǰ / ûл
        if((dist > 0 && (dist < t || matID == 666u))) {
            col += vec3(1.0, 0.95, 0.90)*exp2( -64.0*seg2 );
        }
#else
        // йԴֱӷ
        #define R2 0.01
        // (1) Դǰ / ûл
        // (2) йԴ
        if((dist > 0 && (dist < t || matID == 666u)) && (seg2 < R2)) {
            return vec3( 1.0, 0.95, 0.90 );
        }
#endif
    }

    // postprocess
    col = post(col);

    return col;
}

void main() {
#if AA<2
    vec3 col = computePixel(gl_FragCoord.xy);
#else
    vec3 col = vec3(0.0);
    for( int j = 0; j < AA; j++ ) {
        for( int i = 0; i < AA; i++ ) {
            vec2 of = vec2(i, j) / float(AA);
            col += computePixel(gl_FragCoord.xy + of);
        }
        col /= float(AA*AA);
    }
#endif
    FragColor = vec4(col, 1.0);
}